#include "master_driver.h"
#include "config.h"
#include "move_struct.h"
#include "Gearing.h"

/// cog variables
static int master_driver_cog = 0;
static int master_driver_stack[100];

volatile bool move_complete;

void master_driver(void *par);

void master_driver_init()
{
	if(master_driver_cog == 0)
	{
		master_driver_cog = 1 + cogstart(&master_driver, NULL,
			master_driver_stack, sizeof master_driver_stack);
	}
}

/// stop x_axis cog
void master_driver_stop()
{
	if(master_driver_cog > 0)
	{
		cogstop(master_driver_cog - 1);
		master_driver_cog = 0;
	}
}

void master_driver(void *par)
{
	do
	{
		if(app_move_now == true)
		{
			app_move_now = false;

			int32_t master_counter;

			// Set DIRA as an output.
			DIRA |= 1 << current_move.master_dir_pin;

			// Set up the CTRMODE of Counter A for NCO/PWM single-ended.
			CTRA = (4 << 26) | current_move.master_step_pin;

			// Set the value to be added to PHSA with every clock cycle.
			FRQA = 1;

			// Set DIRA as an output.
			DIRA |= 1 << current_move.master_step_pin;

			// Set the OUTA register to match the desired direction of rotation.
			if(current_move.master_dir == 0)
			{
				OUTA &= (~1 << current_move.master_dir_pin);
			}
			else
			{
				OUTA |= 1 << current_move.master_dir_pin;
			}

			// Get the current System Counter value.
			master_counter = CNT;

			// Ramp up the stepper motor a predetermined amount of steps
			while(current_move.master_ramp_up_steps != 0)
			{
				// Send out a high pulse on the step pin for the desired duration.
				PHSA = -current_move.master_pulse_width;

				// Wait for a specified period of time before sending another
				// high pulse to the step pin.
				waitcnt(master_counter += current_move.master_start_stop_speed
					-= current_move.master_ramp_inc_dec);

				// Decrement ramp_up_steps
				current_move.master_ramp_up_steps--;
			}

			// Run the stepper motor at the current speed for a predetermined amount of steps
			while(current_move.master_running_steps != 0)
			{
				// Send out a high pulse on the step pin for the desired duration.
				PHSA = -current_move.master_pulse_width;

				// Wait for a specified period of time before sending another
				// high pulse to the step pin.
				waitcnt(master_counter += current_move.master_start_stop_speed);

				// Decrement running_steps
				current_move.master_running_steps--;
			}

			// Ramp down the stepper motor a predetermined amount of steps
			while(current_move.master_ramp_down_steps != 0)
			{
				// Send out a high pulse on the step pin for the desired duration.
				PHSA = -current_move.master_pulse_width;

				// Wait for a specified period of time before sending another
				// high pulse to the step pin.
				waitcnt(master_counter += current_move.master_start_stop_speed
					+= current_move.master_ramp_inc_dec);

				// Decrement ramp_down_steps
				current_move.master_ramp_down_steps--;
			}

			move_complete = true;
		}
	}
	while(master_driver_cog > 0);
}